//---------------------------------------------------
// Multidimensional Scaling according to best 
//   distance matrix reconstruction Based on Fisher's z
// 
// Author  : Marc Strickert [hitmds2]  
//           (modified by Max Mignotte)
// module  : MDS.h                                               
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------


#ifndef __MDS_H
#define __MDS_H
//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <math.h>
#include <float.h>
#include <signal.h>

//------------------------------------------------
// TYPEDEF ---------------------------------------
//------------------------------------------------
typedef float FLT;

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define NCHAR 200
#define TROIS 3

#define SCANFMT "%lf"
#define START_ANNEALING_RATIO .5
#define EPS 1e-16

#define frand() ((double)rand()/(RAND_MAX+1.0))
#define irand(X) ((int)((double)(X)* frand()))
#define error(X) fprintf(stderr, (X)), main_bye(), exit(1)
#define ABS(X) (((X) > 0) ? (X) : -(X))

#define D(mat,i,j) (*(mat + ( \
  (i<j) ? (j - 1 + ((((pattern_length<<1) - i - 3) * i) >> 1)) \
        : ((i==j) ? matsize \
                  : (i - 1 + ((((pattern_length<<1) - j - 3) * j) >> 1))) \
    )))

#define inline __inline


//------------------------------------------------
// CLASSE MDS -------------------------
//------------------------------------------------
class MDS
{
  //--Donnees--
  //-----------
  public:

  int correlation_exponent,
    pattern_length,      
    pattern_dimension,   
    matsize,             
    target_dim,          
    printqual,           
    matrix_input;       

  double **pattern,        
         **points,
         points_distmat_mean,
         points_distmat_mixed,
         points_distmat_mono,
         pattern_distmat_var_sum,
         pattern_distmat_mean,
         correps;          

  FLT  *pattern_distmat,
       *points_distmat;

  int *shuffle_index; 
  double mexponent;
  int cycles;
  double rate;

 
  //--Methods--
  //-----------
  public:
  MDS();
  MDS(int);
  ~MDS();

 //>MainMethod
 void ReduceMDS(float**,int,int,float**);

 //>InlineFunctions
 double dist(int,double*,double*);
 double dist4s(int,double*,double*);
 double dist_minkowski(int,double*,double*);
 double dist_spearman(int,double*,double*);
 double mean(int,double*); 
 double meanf(int,FLT*);
 double corr_deriv_vec_flt(int,FLT*,FLT*,FLT*,int,int);
 double corr_deriv_vec(int,double*,double*,double*,int,int);
 double corr_deriv_abs(int,double*,double*); 
 double corr(int,double*,double*); 
 double corr_2();
 double deriv(int,int,int);
 void data_free(int,int);
 int data_alloc();
 
 //>OutlineFunctions
 void My_data_stdin(int,int,int);
 void My_data_init(float**);

 void shuffle_index_free();
 void shuffle_index_alloc();
 void shuffle_do();
 int shuffle_next();
 
 void mds_train(int,double);
 void mds_postprocess();
  
 void analyze_corr_deriv();
 void mds_stdout();
 void mds_stdout2();
 void main_bye();
 void main_bye2();
};

//----------------------------------------------------------
// dist
//---------------------------------------------------------- 
inline
double MDS::dist(int dimension,double *d1,double *d2) 
{
 int u;
 double sum=0.,tmp;

  for(u=0;u<dimension;u++) 
  { tmp = *d1++ - *d2++;
    sum += tmp * tmp; }

  return sqrt(sum);
}

//----------------------------------------------------------
// "quartic" Euclidean 
//---------------------------------------------------------- 
inline
double MDS::dist4s(int dimension,double *d1,double *d2) 
{
 int u;
 double sum=0.,tmp;

  for(u=0;u<dimension;u++) 
    { tmp = *d1++ - *d2++;
      tmp *= tmp;
      sum += tmp*tmp; }

  return sum;
}

//----------------------------------------------------------
// Minkowski metric, mexponent = 2 -> Euclidean distance
//---------------------------------------------------------- 
inline
double MDS::dist_minkowski(int dimension,double *d1,double *d2) 
{
 int u;
 double sum=0.,tmp;
 double mexponent=1.;

 for(u=0;u<dimension;u++) 
   { tmp = *d1++ - *d2++;
     tmp = ABS(tmp);
     sum += pow(tmp,mexponent); }

  return pow(sum,1./mexponent);
}

//----------------------------------------------------------
// Minkowski metric, mexponent = 2 -> Euclidean distance
//---------------------------------------------------------- 
inline
double MDS::dist_spearman(int dimension,double *d1,double *d2) 
{
 static double f=0.;
 static int *idx1,*idx2,dim2,dim4;
 int u,i;
 double sum=0.,tmp,s1=0.,s2=0.;

  if(f==0.) 
    { f=1./(dimension*(dimension*dimension-1.));

      dim2=dimension<< 1; 
      dim4=dim2<<1;

      idx1=(int *)malloc(dim4*sizeof(int));  
      idx2=idx1+dim2;

      for(u=0;u<dim4;u++) idx1[u]=0; }


  /* multiplicity */
  i=0;
  for(u=0;u<dimension;u++) 
     { if(++idx1[(int)(2.*(d1[u]-1.))]>1) ++i;
       if(++idx2[(int)(2.*(d2[u]-1.))]>1) ++i; }

  if(i)  
    for(u=0;u<dim2;u++) 
      { if((i = idx1[u])>1) s1 += i * (i * i - 1);
        if((i = idx2[u])>1) s2 += i * (i * i - 1); }

  for(u=0;u<dimension;u++) 
     { tmp = *d1++ - *d2++;
       sum += tmp*tmp;
       idx1[u] = idx2[u] = idx1[u+dimension] = idx2[u+dimension] = 0; }
  
  if(s1==0. && s2==0.)  return 6.*f * sum;

  else { s2 *= f;
         if((s1 *= f) > 1-EPS) return (s2 > 1-EPS) ? 0 : 1.;
    else if(s2 > 1-EPS) return 1.; }

  return 1.-(1.- 6.*(f*sum+(s1+s2)/12.))/sqrt((1.-s1)*(1.-s2));
}

//----------------------------------------------------------
// mean
//---------------------------------------------------------- 
inline
double MDS::mean(int dimension,double *p) 
{
 int i;
 double sum=0.;

 for(i=0;i<dimension;i++) sum += *p++;
 return sum/dimension;
}

//----------------------------------------------------------
// meanf
//---------------------------------------------------------- 
inline
double MDS::meanf(int dimension,FLT *p) 
{
 int i;
 double sum=0.;
 
 for(i=0;i<dimension;i++) sum += *p++;
 return sum/dimension;
}

//----------------------------------------------------------
// corr_deriv_vec_flt
//---------------------------------------------------------- 
inline 
double MDS::corr_deriv_vec_flt(int dimension,FLT *d1,FLT *d2,FLT *res,int symmetric,int addup) 
{
 int u,isres=(res!=NULL);
 double mono1 = 0., mono2 = 0., mixed = 0., sum = 0., tmp, tmx, f;
 FLT md1=meanf(dimension, d1),md2=meanf(dimension, d2);
 
  for(u=0;u<dimension;u++) 
     { tmx = d1[u] - md1; mono1 += tmx * tmx;
       tmp = d2[u] - md2; mono2 += tmp * tmp;
       mixed +=tmp*tmx; }
  
  tmp=sqrt(mono1*mono2);

  if(tmp < EPS) tmp = EPS;
  tmp = 1. / tmp;

  if(mono1 < EPS) mono1 = EPS;
  mono1 = mixed / mono1;

  if(symmetric) 
    { if(mono2<EPS) mono2=EPS;
      mono2=mixed/mono2; }

  for(u=0;u<dimension;u++) 
    { f=((*d1-md1)-mono2*(*d2- md2))*tmp;

    if(symmetric) 
      { tmx = ((*d2 - md2) - mono1 * (*d1 - md1)) * tmp;
        f *= tmx;
        f = ABS(f); }
    sum += f;

    if(isres) { if(addup) *res++ += f; else *res++ = f; }

    ++d1; ++d2;
  }
 return sum/dimension;  /* average (absolute) contribution */
}

//----------------------------------------------------------
// derivative of Pearson correlation w.r.t d2 or absolute 
// to both directions (symmetric)
//---------------------------------------------------------- 
inline 
double MDS::corr_deriv_vec(int dimension, double *d1, double *d2, double *res, int symmetric, int addup) 
{
 int u,isres=(res!=NULL);
 double mono1=0.,mono2=0.,mixed=0.,sum=0.,tmp,tmx,f;
 double md1=mean(dimension, d1),md2=mean(dimension, d2);
 
  for(u=0;u<dimension;u++) 
     { tmx = d1[u] - md1; mono1 += tmx * tmx;
       tmp = d2[u] - md2; mono2 += tmp * tmp;
       mixed += tmp * tmx; }
  
  tmp=sqrt(mono1 * mono2);

  if(tmp < EPS) tmp=EPS;
  tmp=1./tmp;

  if(mono1<EPS) mono1=EPS;
  mono1=mixed/mono1;

  if(symmetric) 
    { if(mono2<EPS) mono2 = EPS;
      mono2 = mixed / mono2; }

  for(u=0;u<dimension;u++) 
     { f = ((*d1 - md1) - mono2 * (*d2 - md2)) * tmp;

    if(symmetric) 
     { tmx = ((*d2 - md2) - mono1 * (*d1 - md1)) * tmp;
       f *= tmx;
       f = ABS(f); }
    sum += f;

    if(isres) { if(addup) *res++ += f; else *res++ = f; }

    ++d1; ++d2;
     }

  return sum/dimension; 
}

//----------------------------------------------------------
// realize symmetric measure d(x,y) = d(y,x) 
//---------------------------------------------------------- 
inline 
double MDS::corr_deriv_abs(int dimension,double *d1,double *d2) 
 { return corr_deriv_vec(dimension, d1, d2, NULL, 1, 1); }

//----------------------------------------------------------
// (1 - Pearson)^expo correlation
//---------------------------------------------------------- 
inline 
double MDS::corr(int dimension,double *d1,double *d2) 
{
 int u;
 double mono1=0.,mono2=0.,mixed=0.,tmp;
 double md1=mean(dimension,d1),md2 =mean(dimension,d2);

 for(u=0;u<dimension;u++) 
    { tmp = d1[u] - md1; mono1 += tmp * tmp;
      tmp = d2[u] - md2; mono2 += tmp * tmp;
      mixed += tmp * (d1[u] - md1); }
  
 tmp=sqrt(mono1*mono2);
 mixed /= (tmp < EPS) ? EPS : tmp;

  /* allow correlated and anti-correlated patterns to be similar */
  if((u = mexponent < 0.) && mixed < 0.) mixed = -mixed; 
  
  mono1 = u ? -mexponent : mexponent;

  return (mono1 == 1.) ? 1. - mixed : pow(1. - mixed, mexponent);
}

//----------------------------------------------------------
// for distance matrix discrepancy evaluation. Min val: 0
//---------------------------------------------------------- 
inline 
double MDS::corr_2(void)
{
 double p2 = points_distmat_mixed * points_distmat_mixed,e2 = EPS * EPS;

 return pattern_distmat_var_sum*points_distmat_mono/((p2<e2)? e2 : p2) - 1.;
}

//----------------------------------------------------------
// derivative of HiT-MDS stress function
//---------------------------------------------------------- 
inline 
double MDS::deriv(int j, int k, int idx)
{
 static double preres;
 static int _j = -1, _k = -1;
 double d = D(points_distmat,j, k);

 if(_j != j || _k != k) { double  D = D(pattern_distmat,j,k),
                          dif=d-points_distmat_mean;
                          _j=j; 
                          _k=k;
                          preres=dif* points_distmat_mixed-D*points_distmat_mono; }
  
 return preres*(points[j][idx]-points[k][idx])/((d<EPS)?EPS:d);
}

//----------------------------------------------------------
// data_free
//---------------------------------------------------------- 
inline 
void MDS::data_free(int i,int what)
{
  
  while(--i >= 0) 
       { if(pattern != NULL) free(pattern[i]);
         if(what == 0 && points != NULL) free(points[i]); }
  free(pattern),pattern=NULL;
  
  if(what == 0) 
    { free(points), points = NULL;
      free(points_distmat), points_distmat = NULL;
      free(pattern_distmat), pattern_distmat = NULL; }
}

//----------------------------------------------------------
// data_alloc
//---------------------------------------------------------- 
inline 
int MDS::data_alloc(void)
{
 int i, tdim;

 /* also valid for neg values */
 matsize = ((pattern_length-1) * pattern_length) >> 1;

 /* upper triangle matrix */
 tdim = sizeof(FLT) * (1 + matsize);

 if(NULL == (pattern_distmat = (FLT *)malloc(tdim))) error("data_alloc(): cannot malloc pattern_distmat!");
 if(NULL == (points_distmat = (FLT *)malloc(tdim)))  error("data_alloc(): cannot malloc points_distmat!");

 /* init access to diagonal elements: 0 by default */
 D(points_distmat,0,0)=D(pattern_distmat,0,0)=(FLT)0.;
  
 if(pattern_dimension>0) 
   { tdim = sizeof(double) * pattern_dimension;
    
     if(NULL == (pattern = (double **)malloc(sizeof(double *) * pattern_length)))
     error("data_alloc(): cannot malloc pattern!");
    

     for(i = 0; i < pattern_length; i++) 
       {
       if(NULL == (pattern[i] = (double *)malloc(tdim))) 
         { data_free(i,0);
           error("data_alloc(): cannot malloc pattern[k]!"); }
       }
   }
 return 1;
}

#endif 
