//---------------------------------------------------
// module  : SCVOIC.h                         
// auteur  : Ayman Khlif and Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

#ifndef SCVOIC_H
#define SCVOIC_H

//------------------------------------------------
// INCLUDED FILES --------------------------------
//------------------------------------------------
#include "FunctionSCVOIC.h"
#include "Chrono.h"
#include "MDS.h"

//------------------------------------------------
// RESULTS
//------------------------------------------------
//->>BSDS300::Diversity=[3.05]::corr(D,d)=[0.856]

//------------------------------------------------
// CONSTANT --------------------------------------
//------------------------------------------------

//>IMG_BASE
const int LENGTH=321;
const int WIDTH=481;
const int SIZE=MIN(WIDTH,LENGTH);
const int NBSEGS=300;      //>300

//>MODEL_KMEAN_HC_MDS
const int FACTREDSEG=10; 

//>MAPPING_MDS_&_CLUSTERS
const int FACTREDIMGVISU=5;

//>QUERYBYDRAWING
const int NBNEIGHBORS_QBD=3;  

//>DEFAULT_VALUES
const int DIMMAPOUT=2;
const int INIT=2;
const int NBCLASS=8;

//>MODEL_ICM
const int SIZENEIGHB=7; 
const int ITER_ICM=200;

//>VISU
const int SQUA=1;
const int ZOOM=1;

//>USELESS
const int QUIT=0;
const int SAVE=0;

//------------------------------------------------
// CLASS  SCVOIC ---------------------------------
//------------------------------------------------
class SCVOIC
{
  //---Data---
  //----------
  public:

  //Matrix
  int**** VctImgCol;
  int***  VctSeg;
 
  int*    VctLabel;
  int***  VctCtrSeg;

  //Parameters
  int   Width,Length;
  int   NbImg;

  int   NbClass;
  int   DimMap;
  int   Init;

  //Stat
  float WithinClusterInertia;
  float BetweenClusterInertia;
  float FisherDistance;
  float Diversity;
  int   TimeSec;

  //Flag
  int flag_squa;
  int flag_save;

  //Object
  Chrono* chrono;

  //Useful
  char Name_Save[NCHAR],Extension[NCHAR]; 
  char QueryDrawingFileName[NCHAR];
  char BufTmp[NCHAR];

  //--Methods---
  //------------
  public:
  SCVOIC(int);
  ~SCVOIC();

  //>Gestion
  void LoadFiles();
  void ComputeStatClustering(char*);

  //>TestImages
  void FindKNearestNeighbors(float**,int,int,float**);

  //>Methods
  //--------
  void KmeanClusteringSeg();
  void VoiDistMatxForMDS();
  void VoiHierarchicalClustering();  
  void FarestPointAndCenterInformation();  
  void QueryByDrawing();  
  //==================================== 
};

#endif 
