//---------------------------------------------------
// Multidimensional Scaling according to best 
//   distance matrix reconstruction Based on Fisher's z
// 
// Author  : Marc Strickert [hitmds2]  
//           (modified by Max Mignotte)                          
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "MDS.h"

//------------------------------------------------
// CONSTRUCTEUR & DESTRUCTEUR DE LA CLASSE MDS  --
//------------------------------------------------
//----------------------------------------------------------
// Constructeur-1- de la classe MDS 
//----------------------------------------------------------
MDS::MDS()           
 { 
  //>ParamConvergence
  cycles=1000;
  rate=10.0;

  //>Presentation
  printf("\n\n  >> MultidimensionalScaling");
  fflush(stdout);
 
  //general Minkowski metric, mexponent = 2 
  //-> Euclidean distance 
  mexponent=1.;
 }

//----------------------------------------------------------
// Constructeur-2- de la classe MDS 
//----------------------------------------------------------
MDS::MDS(int cyc)           
 { 
  //>ParamConvergence
  cycles=cyc;
  rate=10.0;

  //>Presentation
  printf("\n\n   >> MultidimensionalScaling");
  fflush(stdout);
 
  //general Minkowski metric, mexponent = 2 
  //-> Euclidean distance 
  mexponent=1.;
 }

//----------------------------------------------------------
// Destructeur de la classe MDS
//----------------------------------------------------------
MDS::~MDS()
{ ;  }


//------------------------------------------------
// ReduceMDS
//------------------------------------------------
void MDS::ReduceMDS(float** MyDistxMatx,int NbData,int DimTarget,float** RedData)
{
 int i,j;

  //>Presentation
 printf(" [-> To Dim %d]\n",DimTarget);

  //>Treatment
  My_data_stdin(NbData,0,-DimTarget);
  My_data_init(MyDistxMatx);
  printf("# corr(D,d): %g\n",1./sqrt(corr_2()+1));
  mds_train(cycles*pattern_length,rate);
  main_bye2();

  //>Recup
  for(i=0;i<NbData;i++) for(j=0;j<DimTarget;j++) RedData[i][j]=points[i][j];

  //>FreeMemory
  shuffle_index_free();
}

//------------------------------------------------
// GESTION  -------------------------------------
//------------------------------------------------
//------------------------------------------------
// My_data_stdin
//------------------------------------------------
void MDS::My_data_stdin(int pat_lgth,int pat_dim,int tar_dim)
{
 double *random_matrix=NULL;
 int line,tdim,column,j,off=0;

 pattern_length=pat_lgth;
 pattern_dimension=pat_dim;
 target_dim=tar_dim;

 if(!data_alloc()) error("main(): data_alloc-Error!");
 matrix_input=0;
  
 if((tdim=target_dim)<0) 
   {  
    target_dim=-target_dim;
    if(!matrix_input) 
      { j=target_dim*pattern_dimension; 
        if(NULL==(random_matrix=(double *)malloc(sizeof(double)*j)))
        error("data_alloc(): cannot malloc projection matrix!");      
        while(--j>=0) random_matrix[j]=2.*frand()-1.; }
   }
  
  if(NULL == (points = (double **)malloc(sizeof(double *) * pattern_length)))
    error("data_alloc(): cannot malloc points!");
  
  for(line=0;line<pattern_length;line++,off++) 
  { if(NULL==(points[line]=(double *)malloc(sizeof(double)* target_dim))) 
      { data_free(line,0);
        error("data_alloc(): cannot malloc points[k]!"); }  
    for(column=0;column<target_dim;column++) points[line][column]=2.*frand()-1.;  }

 /* Info */
  if (0)
  {
  printf("\n [tdim=%d]",tdim);
  for(line=0;line<pattern_length;line++) 
    { printf("\n");
      for(column=0;column<target_dim;column++)
	printf("<%.2f>",points[line][column]); }
  }

  free(random_matrix);
}

//----------------------------------------------------------
// MyDistMatx[0][1]                                         
//           [0][2]                                         
//           [0][3]                                         
//           [1][2]                                        
//           [1][3]                                         
//           [2][3]                                         
//----------------------------------------------------------
void MDS::My_data_init(float** MyDistMatx)
{
 double tmp;
 int i,j;

  /* fill distance matrices */
  for(i=0;i<pattern_length-1;i++) for(j=i+1;j<pattern_length;j++) 
     { D(pattern_distmat,i,j)=MyDistMatx[i][j];    
       D(points_distmat,i,j)=(FLT)dist(target_dim,points[i],points[j]); }

  /* output distance */
  if (0)
  {
  printf("(\n");
  for(i=0;i<matsize;i++) printf("[%g]\n",pattern_distmat[i]);
  printf(")\n");
  }

  /* initial values of point distance matrix, mean, mixed, mono */
  points_distmat_mean=meanf(matsize,points_distmat);
  
  /* mean value of pattern distance matrix will be subtracted */
  pattern_distmat_mean=meanf(matsize,pattern_distmat);

  points_distmat_mixed=points_distmat_mono=pattern_distmat_var_sum=0.;

  for(i=0;i<matsize;i++) 
   { tmp=points_distmat[i]-points_distmat_mean;
     points_distmat_mixed += tmp*(pattern_distmat[i]-=pattern_distmat_mean);
     points_distmat_mono +=  tmp*tmp;
     pattern_distmat_var_sum += pattern_distmat[i]*pattern_distmat[i]; }
}


/******************************************************************************/
/*              Some data shuffling routines                                  */
/******************************************************************************/

#define SHUFFLE_NUMBER pattern_length
//------------------------------------------------
// shuffle_index_free
//------------------------------------------------
void MDS::shuffle_index_free(void)
{ free(shuffle_index); }

//------------------------------------------------
// shuffle_index_alloc
//------------------------------------------------
void MDS::shuffle_index_alloc(void)
{
 int len=SHUFFLE_NUMBER;

  shuffle_index_free();

  if((shuffle_index = (int *)malloc(sizeof(int) * len)) == NULL)
    error("shuffle_index_alloc(): cannot malloc() shuffle_index");

  while(--len >= 0) shuffle_index[len] = len;
}

//------------------------------------------------
// shuffle_do
//------------------------------------------------
void MDS::shuffle_do()
{
 int cnt,ind,tmp;

 if(shuffle_index == NULL) shuffle_index_alloc();

 for(cnt=0;cnt<SHUFFLE_NUMBER;cnt++) 
   { ind = irand(SHUFFLE_NUMBER);
     tmp = shuffle_index[cnt];
     shuffle_index[cnt] = shuffle_index[ind];
     shuffle_index[ind] = tmp; }
}

//------------------------------------------------
// shuffle_nex
//------------------------------------------------
int MDS::shuffle_next(void)
{
 static int ind = INT_MAX-1;

 if(++ind >= SHUFFLE_NUMBER) 
   { shuffle_do();
     ind=0; }

 return shuffle_index[ind];
}


//------------------------------------------------
// mds_train
//------------------------------------------------
void MDS::mds_train(int cycles,double learning_rate)
{
 double *delta_point,*point,dtmp,diff,diff_mixed,diff_mono,*diffs,diffs_mean;
 FLT *ptmp;
 int c, k, i, j, m, t;

 if(NULL == (delta_point = (double*)malloc(target_dim * sizeof(double))))
 error("mds_train(): cannot malloc() helper point!");

 if(NULL == (diffs = (double*)malloc(pattern_length * sizeof(double))))
 error("mds_train(): cannot malloc() diffs vector!");


  t=0,m=cycles/10;

  for(c=0;c<cycles;c++) 
  {
    
    if(++t == m) {
      t = 0;
      fprintf(stderr, "%3.2f%%: %g  \t(r = %g)\n", 
                         100. * c/cycles, corr_2(), 
                         1./sqrt(corr_2()+1.));
    }
    
    
    i = shuffle_next();
    point = points[i];

    
    for(k = 0; k < target_dim; k++) delta_point[k] = 0.;
    
    for(j = 0; j < pattern_length; j++) {
      
      if(j != i) {
        
        for(k = 0; k < target_dim; k++) {
          
          delta_point[k] += deriv(i, j, k);

        } 
      }
    }
    

    if((diff_mixed = START_ANNEALING_RATIO * cycles - c) < 0.)  {
      diff_mixed = learning_rate * (1. + 1. / (1.-START_ANNEALING_RATIO) * diff_mixed / cycles);
    }else
      diff_mixed = learning_rate;
    
    for(k = 0; k < target_dim; k++) {
      diff = diff_mixed * delta_point[k] / ABS(delta_point[k]);
      delta_point[k] = point[k] - diff /* * (2. * frand() - .5) */;
    }


    /* track change of mean, mixed, mono { */
    diff = 0.;  
    
    for(k = 0; k < pattern_length; k++) 
      if(k != i) {
        dtmp = dist(target_dim, delta_point, points[k]);
        diff += (diffs[k] = dtmp - D(points_distmat, i, k));
      }
    
    diffs_mean = diff / matsize;
    dtmp = -diffs_mean - points_distmat_mean;
    diff_mixed = diff_mono = 0.;

    for(k = 0; k < pattern_length; k++) 
      if(k != i) {
        ptmp = &D(points_distmat, i, k);        
        diff_mixed += D(pattern_distmat, i, k) * diffs[k];
        diff_mono += diffs[k] * (diffs[k] + 2. * (*ptmp + dtmp));        
        *ptmp += diffs[k]; /* ugly */
      }

    points_distmat_mean += diffs_mean;
    points_distmat_mixed += diff_mixed;   
    points_distmat_mono += diff_mono + matsize * diffs_mean * diffs_mean;

    /* } track changes of mean, mixed, mono */

    for(k = 0; k < target_dim; k++) 
       point[k] = delta_point[k];

  } /* for cycles */

  free(diffs);
  free(delta_point);
}

//------------------------------------------------
// mds_postproces
// move center to origin, scale data in such a way that 
// standard deviation of dimension of largest variance is equal to 1 
//------------------------------------------------
void MDS::mds_postprocess(void) 
{
 double mean, var = 1., maxvar = -1.;
 int i, j;
  
 for(j = 0; j < target_dim; j++) 
   { mean=0.;
     for(i=0;i<pattern_length;i++) mean += points[i][j];
     mean /= pattern_length;
     for(i=0;i<pattern_length;i++) points[i][j] -= mean; }

  for(j=0;j<target_dim;j++) 
    { var=0.;
      for(i=0;i<pattern_length;i++) var += points[i][j]*points[i][j];
      var /= pattern_length - 1;
      if(var > maxvar) maxvar = var; }

  var=1./sqrt(maxvar);  
  for(j=0;j<target_dim;j++) for(i=0;i<pattern_length;i++) points[i][j]*=var;
}

//------------------------------------------------
// analyze_corr_deriv
//------------------------------------------------
void MDS::analyze_corr_deriv() 
{
 int i, j;
 double *derivs;

  if(NULL == (derivs = (double *)malloc(pattern_dimension)))
    error("analyze_corr_deriv(): cannot malloc deriv[] vector!");

  for(i=0;i<pattern_dimension;i++) derivs[i]=0.;

  for(i=0;i<pattern_length-1;i++) 
     { for(j=i+1;j<pattern_length;j++) 
         { /* fprintf(stderr, "%d %d %g\n", i, j, corr(pattern_dimension, pattern[i], pattern[j])); */
           corr_deriv_vec(pattern_dimension, pattern[i],pattern[j],derivs,1,1); } }

  for(i=0;i<pattern_dimension;i++) printf("%g\n", derivs[i]/(pattern_length*(pattern_length-1)));

  free(derivs);
}

//------------------------------------------------
// mds_stdout
//------------------------------------------------
void MDS::mds_stdout(void) 
{
 int i,j;
 double sum=0.,tmp;

 if(printqual) {  /* correlation influence w.r.t. points_distmat */ 
                corr_deriv_vec_flt(matsize,pattern_distmat,points_distmat,points_distmat,0,0); }
  
 for(i=0;i<pattern_length;i++) 
  {
    if(printqual) 
      { sum = 0.;
        for(j=0;j<pattern_length;j++) 
           { tmp = D(points_distmat,i,j);
             sum += ABS(tmp); } }

    for(j=0;j<target_dim;j++) printf("%g  ",points[i][j]);

    if(printqual) printf("%g\n",sum/(pattern_length-1));
    else  printf("\n");
  }
 fflush(stdout);
}

//------------------------------------------------
// mds_stdout2
//------------------------------------------------
void MDS::mds_stdout2(void) 
{
 if(printqual) {  /* correlation influence w.r.t. points_distmat */ 
                corr_deriv_vec_flt(matsize,pattern_distmat,points_distmat,points_distmat,0,0); }
}

//------------------------------------------------
// main_bye
//------------------------------------------------
void MDS::main_bye() 
{
  printf("# corr(D,d): %g\n", 1./sqrt(corr_2()+1));
  mds_postprocess();
  mds_stdout();
  shuffle_index_free();
  data_free(pattern_length,0);
}

//------------------------------------------------
// main_bye2
//------------------------------------------------
void MDS::main_bye2() 
{
  printf("# corr(D,d): %g\n", 1./sqrt(corr_2()+1));
  mds_postprocess();
  mds_stdout2();
}
