//---------------------------------------------------
// module  : f_SCVOIC.cc                       
// auteur  : Ayman Khlif and Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "SCVOIC.h"

//------------------------------------------------
// METHODES DE LA CLASSE SCVOIC  ------------------
//------------------------------------------------
//----------------------------------------------------------
// Constructeur de la classe  SCVOIC
//----------------------------------------------------------
SCVOIC::SCVOIC(int NbCl)           
 {
 //>InitPar
 NbImg=NBSEGS;
 NbClass=NbCl;

 //>MemoryAlloc
 VctLabel=imatrix_allocate_1d(NbImg);

 //>StackSizeIncreased(128 MB)
 const rlim_t kStackSize=128*1024*1024;   
 struct rlimit rl;
 int result=getrlimit(RLIMIT_STACK,&rl);
 if (result==0)
 if (rl.rlim_cur<kStackSize)
    {  rl.rlim_cur=kStackSize;
       result=setrlimit(RLIMIT_STACK,&rl);
       if (result!=0) fprintf(stderr,"setrlimit returned result = %d\n",result); } 

 //>InitFlag
 flag_squa=0;
 flag_save=0;
 }

//----------------------------------------------------------
// Destructeur de la classe VOIFBM
//----------------------------------------------------------
SCVOIC::~SCVOIC()
{ 
 //>Time
 TimeSec=(*chrono).elapsed(SECOND); (*chrono).print();
 fflush(stdout);

 //>FreeMemory 
 if (VctSeg)      free_imatrix_3d(VctSeg,NbImg);
 if (VctImgCol)   free_imatrix_4d(VctImgCol,NbImg,TROIS);
 if (VctLabel)    free_imatrix_1d(VctLabel);
 if (VctCtrSeg)   free_imatrix_3d(VctCtrSeg,NbClass);
}

//---------//
//-GESTION-//
//---------//
//----------------------------------------------------------
// LoadFiles [Imgs_&_Segs] & VctCtrSeg_Alloc
//----------------------------------------------------------
void SCVOIC::LoadFiles()
{
 //>TimeObjet
 chrono=new Chrono;
 (*chrono).reset();

 //>Img>SQUARE_DIMENSION
 if (flag_squa)
   {  printf("\n  >> [[ Squared Image ]]");
      VctSeg=imatrix_allocate_3d(NbImg,SIZE,SIZE);
      VctImgCol=imatrix_allocate_4d(NbImg,TROIS,SIZE,SIZE); 
      LoadTabSegSqAndImg(VctImgCol,VctSeg,NbImg,LENGTH,WIDTH,1);
      Width=MIN(LENGTH,WIDTH);
      Length=Width; }


 if (!flag_squa)
   {  printf("\n   >> [[ Image (%d,%d) ]]",LENGTH,WIDTH);
      VctSeg=imatrix_allocate_3d(NbImg,LENGTH,WIDTH);
      VctImgCol=imatrix_allocate_4d(NbImg,TROIS,LENGTH,WIDTH); 
      LoadTabSegAndImg(VctImgCol,VctSeg,NbImg,LENGTH,WIDTH,1);
      Width=WIDTH;
      Length=LENGTH; }

 //>MemoryAlloc
 VctCtrSeg=imatrix_allocate_3d(NbClass,Length,Width);
}

//----------------------------------------------------------
// ComputeStatClustering
// [VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg]
//
//----------------------------------------------------------
void SCVOIC::ComputeStatClustering(char* Name)
{
 int k;

 //>Stat
 WithinClusterInertia=WithinInertieVOI(VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg);
 BetweenClusterInertia=BetweenInertieVOI(Length,Width,NbClass,VctCtrSeg);
 FisherDistance=WithinClusterInertia/BetweenClusterInertia;
 printf("\n\n Within_Cluster_Inertia(NbSegs=%d)=[%.5f]",NbImg,WithinClusterInertia);
 printf("\n Between_Cluster_Inertia(NbClass=%d)=[%.5f]",NbClass,BetweenClusterInertia);
 printf("\n FisherDistance=[%.5f]",FisherDistance);

 printf("\n\n ---CLASSES:::\n");
 for(k=0;k<NbImg;k++) printf("[%d::cl=%d]",k,VctLabel[k]);

 strcpy(Name_Save,Name);
 sprintf(BufTmp,"_Cl%d",NbClass);
 strcat(Name_Save,BufTmp);
 sprintf(BufTmp,"_F%.3f",FisherDistance);
 strcat(Name_Save,BufTmp);
 sprintf(BufTmp,"_N%d",NbImg);
 strcat(Name_Save,BufTmp);
 sprintf(BufTmp,"_P%d_",Init);
 strcat(Name_Save,BufTmp);
 if (!flag_squa) { sprintf(BufTmp,"_NSQ"); strcat(Name_Save,BufTmp); }
 printf("\n\n SaveFileName>[%s]",Name_Save);
}

//--------------------------//
//---POUR TESTS ET IMAGES---//
//--------------------------//
//----------------------------------------------------------
// FindKNearestNeighbors
//----------------------------------------------------------
void SCVOIC::FindKNearestNeighbors(float** MatxDistVOI,int NbNeighbors,int NbCand,float** ReducedData)
{
 int i,j,k,l,p,m;
 int number_cand,number_img,number_seg;
 float min,minX,minY;
 char Buftmp[NCHAR];

 //>MemoryAllocation
 float** MatxDistVOITmp=fmatrix_allocate_2d(NbImg,NbImg);
 int**  ListCand=imatrix_allocate_2d(NbCand,NbNeighbors+1);
 float*** ListPts=fmatrix_allocate_3d(2,NbCand,NbNeighbors+1);
 int*** VctSeg__=imatrix_allocate_3d(NbNeighbors+1,Length,Width);
 int**** VctImgCol__=imatrix_allocate_4d(NbNeighbors+1,TROIS,Length,Width);
 float** ReducedData__=fmatrix_allocate_2d(NbNeighbors+1,2);

 //Init
 for(k=0;k<NbImg;k++)  for(l=0;l<NbImg;l++) MatxDistVOITmp[k][l]=MatxDistVOI[k][l];
 for(k=0;k<NbImg;k++)  MatxDistVOITmp[k][k]=100.0;
 for(k=0;k<NbCand;k++) for(l=0;l<NbNeighbors;l++) ListCand[k][l]=-1;
 
 //>SeekDistMin____________________________________
 //NbCand=1;
 for(p=0;p<NbCand;p++)
    { 
     min=BIGNUMBER;
     number_cand=-1;
     for(k=0;k<NbImg;k++) for(l=0;l<NbImg;l++)
     if (MatxDistVOITmp[k][l]<min) { min=MatxDistVOITmp[k][l]; number_cand=k; }

     //>SelectImg                  
                                 //p=1 //Serpent&criquet
     if (p==1) number_cand=206;  //p=2 //5Dauphins 
     if (p==2) number_cand=193;  //p=3 //1Pyramide 
     if (p==3) number_cand=18;   //p=4 //TigresDansEau    
     if (p==4) number_cand=208;  //p=5 //3Astronautes
     if (p==5) number_cand=171;  //p=6 //3Pyramides

     ListCand[p][0]=number_cand;
 
     for(m=0;m<NbNeighbors;m++)
        { min=BIGNUMBER;
          number_img=-1;
          for(l=0;l<NbImg;l++) if (MatxDistVOITmp[number_cand][l]<min) 
	    { min=MatxDistVOITmp[number_cand][l]; number_img=l; }
    
          if ((number_cand>=0)&&(number_img>=0)) 
	     { MatxDistVOITmp[number_cand][number_img]=100.0;
               MatxDistVOITmp[number_img][number_cand]=100.0; }

          ListCand[p][m+1]=number_img;

          for(l=0;l<NbImg;l++) 
             { MatxDistVOITmp[number_img][l]=100.0;  
               MatxDistVOITmp[l][number_img]=100.0; } }

          for(l=0;l<NbImg;l++) 
             { MatxDistVOITmp[number_cand][l]=100.0;  
               MatxDistVOITmp[l][number_cand]=100.0; } 
    }//_______________________________________________

 //>FillPts
 for(p=0;p<NbCand;p++) for(m=0;m<=NbNeighbors;m++)
       { number_cand=ListCand[p][m];
         ListPts[0][p][m]=ReducedData[number_cand][0];
         ListPts[1][p][m]=ReducedData[number_cand][1]; }

 //>Normalisation
 for(p=0;p<NbCand;p++)
   { minX=ListPts[0][p][0];
     minY=ListPts[1][p][0];
     for(m=0;m<=NbNeighbors;m++)
        { if (ListPts[0][p][m]<minX) minX=ListPts[0][p][m];
          if (ListPts[1][p][m]<minY) minY=ListPts[1][p][m]; }

     for(m=0;m<=NbNeighbors;m++)
        { ListPts[0][p][m]-=minX;
          ListPts[1][p][m]-=minY; } } 

 //>Fill__VctSeg__VctImgCol__Visu__
 for(k=0;k<NbCand;k++)
    {
      printf("\n  [%d]>",k);
     
      for(l=0;l<=NbNeighbors;l++)
	 { number_seg=ListCand[k][l];
           for(i=0;i<Length;i++) for(j=0;j<Width;j++) 
               VctSeg__[l][i][j]=VctSeg[number_seg][i][j];
 
           for(m=0;m<TROIS;m++) for(i=0;i<Length;i++) for(j=0;j<Width;j++)
	       VctImgCol__[l][m][i][j]=VctImgCol[number_seg][m][i][j]; 

           ReducedData__[l][0]=powf(1+ListPts[0][k][l],0.01);
           ReducedData__[l][1]=powf(1+ListPts[1][k][l],0.01); 

           printf("[%d>%.4f:%.4f]",ListCand[k][l],ReducedData__[l][0],ReducedData__[l][1]); }

      sprintf(Buftmp,"%d__",k+1);
      Visu2D_Img(VctImgCol__,NbNeighbors+1,Length,Width,Buftmp);

      //>Other
      //if (k==0) Visu2DMapppingMDS_Img_E(VctImgCol__,NbNeighbors+1,Length,Width,ReducedData__,3,1.0,1.0,0.00001);
    }

 //>Visu
 //for(k=0;k<NbCand;k++) 
 //  { printf("\n  [%d]>",k);
 //    for(l=0;l<=NbNeighbors;l++) printf("[%d>%.4f:%.4f]",ListCand[k][l],ListPts[0][k][l],ListPts[1][k][l]); }
  
 //>FreeMemory
 if (MatxDistVOITmp) free_fmatrix_2d(MatxDistVOITmp);
 if (ListCand) free_imatrix_2d(ListCand);
 if (ListPts)  free_fmatrix_3d(ListPts,2);
 if (VctSeg__)      free_imatrix_3d(VctSeg__,NbNeighbors+1);
 if (VctImgCol__)   free_imatrix_4d(VctImgCol__,NbNeighbors+1,TROIS);
 if (ReducedData__) free_fmatrix_2d(ReducedData__); 
}

//-------------------------------------------//
//---CLUSTERING-SEGMENTATIONS  <METHODS>-----//
//-------------------------------------------//
//----------------------------------------------------------
// [1]>KmeanClusteringSeg
// ----------------------
//
// [opt=0]> Voibfm_ :: << Init: Lines_Hor_Vert >>
// [opt=1]> IMpossible << Init: Lines_Hor_Vert LowRes>HighRes >> 
// [opt=2]> VoibfmM ::  << Init: Best of NSEGS >>
//
//----------------------------------------------------------
void SCVOIC::KmeanClusteringSeg()
{
 int k;   

 //Init
 if (Init==1) Init=2;

 //>Presentation
 printf("\n  VOI-K-MeanClusteringSeg [NbCl:%d][Opt:-t:%d][FactRedSeg=%d]",NbClass,Init,FACTREDSEG);
 printf("\n  -----------------------------------------------------------");
 if (Init==0) printf("\n  << Init: Lines_Hor_Vert >>");
 if (Init==2) printf("\n  << Init: Best of NSEGS  >>");

 //>LoadFiles
 LoadFiles();

 //>MemoryAllocation
 int*** VctSegRed=imatrix_allocate_3d(NbImg,Length/FACTREDSEG,Width/FACTREDSEG);
 int*** VctCtrSegRed=imatrix_allocate_3d(NbClass,Length/FACTREDSEG,Width/FACTREDSEG);

 //>DownSampling
 DownSamplingVctSeg(VctSeg,NbImg,Length,Width,FACTREDSEG,VctSegRed);

 //--------------------------------
 //>Multires-Kmean-----------------
 //--------------------------------
 KmeansVOI(VctSegRed,NbImg,Length/FACTREDSEG,Width/FACTREDSEG,NbClass,VctLabel,VctCtrSegRed,Init);
 //>DEBUG
 //VisuTabSegAll(VctSegRed,NbImg,Length/FACTREDSEG,Width/FACTREDSEG,VctLabel,NbClass,VctCtrSegRed,1,"ClusterCenterLowRes");
 //exit(-1);
 for(k=0;k<NbClass;k++) UpSampling(VctCtrSegRed[k],VctCtrSeg[k],FACTREDSEG,Length,Width);
 //>DEBUG
 //Visu2D_Img(VctCtrSeg,NbClass,Length,Width,"Img_1");
 //exit(-1);
 ComputeCtr(VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg,3);
 //>DEBUG
 //Visu2D_Img(VctCtrSeg,NbClass,Length,Width,"Img_2");
 //exit(-1);
 //--------------------------------

 //>Stat
 ComputeStatClustering((char*)"VoiKMeanClusteringSeg");

 //>Visu_&_Save_Clusters_&_Prototypes
 VisuTabSegAll_Img(VctImgCol,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save);
 VisuTabSegAll_Seg(VctImgCol,VctSeg,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save); 

 //>FreeMemory
 if (VctSegRed)    free_imatrix_3d(VctSegRed,NbImg);
 if (VctCtrSegRed) free_imatrix_3d(VctCtrSegRed,NbClass);
}

//----------------------------------------------------------
// [2]>VoiDistMatxForMDS
// ---------------------
//
// <opt=0> << Init: Lines_Hor_Vert LowRes>HighRes >>
// <opt=1> << Init: Best of NSEGS LowRes>HighRes >>
// <opt=2> << Init: Best of NSEGS >>
//----------------------------------------------------------
void SCVOIC::VoiDistMatxForMDS()
{
 int k,l;

 //>Presentation
 printf("\n  VoiDistMatxForMDS [NbCl:%d][opt:%d][FactRedSeg=%d]",NbClass,Init,FACTREDSEG);
 printf("\n  ------------------------------------------------");
 if (Init==0) printf("\n  << Init: Lines_Hor_Vert L>H >>");
 if (Init==1) printf("\n  << Init: Best of NSEGS L>H >>");
 if (Init==2) printf("\n  << Init: Best of NSEGS >>");

 //>LoadFiles
 LoadFiles();

 //>Normalize-Useless
 //Normalize(VctSeg,NbImg,Length,Width);

 //>MemoryAllocation
 float** MatxDistVOI=fmatrix_allocate_2d(NbImg,NbImg);
 float** ReducedData=fmatrix_allocate_2d(NbImg,DimMap);
 float** TabRange=fmatrix_allocate_2d(DimMap,DIMMAPOUT);
 int*** VctSegRed=imatrix_allocate_3d(NbImg,Length/FACTREDSEG,Width/FACTREDSEG);

 //>SubSampling
 DownSamplingVctSeg(VctSeg,NbImg,Length,Width,FACTREDSEG,VctSegRed);

 //>VoI-DistMatx
 CompMatxDistVOI(VctSegRed,NbImg,Length/FACTREDSEG,Width/FACTREDSEG,MatxDistVOI);
 Diversity=ComputeDiversity(MatxDistVOI,NbImg);
 printf(":: Diversity=[%.2f]",Diversity);

 //>DEBUG
 //VisuMatrix(MatxDistVOI,NbImg,NbImg);
  
 //--------------------------------------------------
 //>MDS-------------
 //--------------------------------------------------
 for(k=0;k<NbImg;k++) for(l=0;l<DIMMAPOUT;l++) ReducedData[k][l]=0.0;
 printf("\n[NBSEGS=%d][PARAM1=%d]",NbImg,DIMMAPOUT);
 fflush(stdout);
 MDS* mds=new MDS();
 (*mds).ReduceMDS(MatxDistVOI,NbImg,DIMMAPOUT,ReducedData);
 RangeAlwaysPositive(ReducedData,NbImg,DimMap);
 //------------------------------------------------

 //>FOR-TESTS-AND-PICTURES-IN-PAPER
 //FindKNearestNeighbors(MatxDistVOI,6,9,ReducedData); 

 //>DEBUG
 //for(k=0;k<NbImg;k++)
 //{ printf("\n"); for(l=0;l<DimMap;l++) printf("[%.2f]",ReducedData[k][l]); }
 
 //>Stat-MDS
 ComputeStat(ReducedData,NbImg,DimMap,TabRange);
 printf("\n  >> Range::[Xmin::Xmax::Ymin:Ymax]::");
 for(k=0;k<DimMap;k++) for(l=0;l<2;l++) printf("[%.3f]",TabRange[k][l]);

 //>Visu-MDS 
 Visu2DMapppingMDS_Seg(VctImgCol,VctSeg,NbImg,Length,Width,ReducedData,FACTREDIMGVISU,6.0);
 Visu2DMapppingMDS_Img(VctImgCol,NbImg,Length,Width,ReducedData,FACTREDIMGVISU,6.0);  
 //Visu2DMapppingMDS_Img_E(VctImgCol,NbImg,Length,Width,ReducedData,6,0.75,1.8,1.2); 
 //exit(-1);

 //>Clustering
 KmeanFast(ReducedData,NbClass,NbImg,DimMap,VctLabel,0,0);
 ComputeCtr(VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg,Init);

 //>Stat
 ComputeStatClustering((char*)"VoiDistMatxForMDS");

 //>Visu_&_Save_Clusters_&_Prototypes
 //VisuTabSegAll_Img(VctImgCol,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save);
 VisuTabSegAll_Seg(VctImgCol,VctSeg,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save); 

 //>FreeMemory
 if (MatxDistVOI) free_fmatrix_2d(MatxDistVOI);
 if (ReducedData) free_fmatrix_2d(ReducedData);
 if (TabRange)    free_fmatrix_2d(TabRange);
 if (VctSegRed)   free_imatrix_3d(VctSegRed,NbImg);
}

//----------------------------------------------------------
// [3]>HierarchicalClustering
// --------------------------
//
// <opt=0> << Init: Lines_Hor_Vert LowRes>HighRes >>
// <opt=1> << Init: Best of NSEGS LowRes>HighRes >>
// <opt=2> << Init: Best of NSEGS >>
//----------------------------------------------------------
void SCVOIC::VoiHierarchicalClustering()
{
 //>Presentation
 printf("\n  VoiHierarchicalClustering [opt:%d][FactRedSeg=%d]",Init,FACTREDSEG);   
 printf("\n  -------------------------------------------------");
 if (Init==0) printf("\n  << Init: Lines_Hor_Vert L>H >>");
 if (Init==1) printf("\n  << Init: Best of NSEGS L>H >>");
 if (Init==2) printf("\n  << Init: Best of NSEGS >>");

 //>LoadFiles
 LoadFiles();

 //>MemoryAllocation
 int*** TabMatxCent=imatrix_allocate_3d(NbImg,Length,Width);
 int**  MatxDistFus=imatrix_allocate_2d(NbImg,NbImg);

 //----------------------------------------------
 //>HierarchicalClustering-------------
 //----------------------------------------------
 HierarCluster(VctSeg,NbImg,Length,Width,TabMatxCent,MatxDistFus,FACTREDSEG,Init);
 VisuHierCl(VctImgCol,VctSeg,NbImg,Length,Width,TabMatxCent,MatxDistFus,Init);
 SeekNbClusters(VctLabel,NbClass,NbImg,MatxDistFus);
 ComputeCtr(VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg,Init);
 //----------------------------------------------

 //>Stat
 ComputeStatClustering((char*)"VoiHierClust");

 //>Visu_&_Save_Clusters_&_Prototypes
 //VisuTabSegAll_Img(VctImgCol,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save);
 VisuTabSegAll_Seg(VctImgCol,VctSeg,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save);

 //>FOR-TESTS___________________________
 if (0)
    { 
     int NbClassMax=16;
     if (VctCtrSeg) free_imatrix_3d(VctCtrSeg,NbClass);
     VctCtrSeg=imatrix_allocate_3d(NbClassMax,Length,Width);

     for(int k=2;k<NbClassMax;k++)
       { printf("\n  <<<<[[%d]]>>>>",k);
          NbClass=k;
          SeekNbClusters(VctLabel,NbClass,NbImg,MatxDistFus);
          ComputeCtr(VctSeg,NbImg,Length,Width,NbClass,VctLabel,VctCtrSeg,Init);
          ComputeStatClustering((char*)"VoiHieraClust");
          VisuTabSegAll_Seg(VctImgCol,VctSeg,NbImg,Length,Width,VctLabel,NbClass,VctCtrSeg,FACTREDIMGVISU,Name_Save); } 

     if (VctCtrSeg) free_imatrix_3d(VctCtrSeg,NbClassMax);
    }
 //_____________________________________

 //>FOR-TESTS___________________________
 if (0)
    { 
    VisuAgglomerations(MatxDistFus,VctImgCol,30,Length,Width,FACTREDIMGVISU,Name_Save);
    }
 //_____________________________________
 
 //>FreeMemory
 if (MatxDistFus) free_imatrix_2d(MatxDistFus);
 if (TabMatxCent) free_imatrix_3d(TabMatxCent,NbImg);
}


//----------------------------------------------------------
// [4]>FarestPointAndCenterInformation
// -----------------------------------
//
// <opt=0> << Init: Lines_Hor_Vert LowRes>HighRes >>
// <opt=1> << Init: Best of NSEGS LowRes>HighRes >>
// <opt=2> << Init: Best of NSEGS >>
//----------------------------------------------------------
void SCVOIC::FarestPointAndCenterInformation()
{
 int k;
 int IndxBestSeg,IndxMaxVoI;;
 float MaxVoI,VoI;

 //>Init
 Init=0;

 //>Presentation
 printf("\n  FarestPointAndCenterInfo [opt:%d][FactRedSeg=%d]",Init,FACTREDSEG);   
 printf("\n  ------------------------------------------------");
 if (Init==0) printf("\n  << Init: Lines_Hor_Vert L>H >>");
 if (Init==1) printf("\n  << Init: Best of NSEGS L>H >>");
 if (Init==2) printf("\n  << Init: Best of NSEGS >>");

 //>LoadFiles
 LoadFiles();

 //>MemoryAllocation
 float*  VctVoIMoy=fmatrix_allocate_1d(NbImg);
 float*  VctVoIDist=fmatrix_allocate_1d(NbImg);
 int**   MatSegCenter=imatrix_allocate_2d(Length,Width);
 int***  MatImgCenter=imatrix_allocate_3d(TROIS,Length,Width);
 int**   MatSegFarest=imatrix_allocate_2d(Length,Width);
 int***  MatImgFarest=imatrix_allocate_3d(TROIS,Length,Width);
 int**   MatSegVoIMoy=imatrix_allocate_2d(Length,Width);

 //>CenterPrototypeImg
 IndxBestSeg=SelectSegMap(VctSeg,Length,Width,NbImg,VctVoIMoy);

 CopyMat(MatSegCenter,VctSeg[IndxBestSeg],Length,Width);
 SaveImagePgm((char*)"MatSegCtr","",MatSegCenter,Length,Width);
 CopyMat(MatImgCenter,VctImgCol[IndxBestSeg],Length,Width);
 SaveImagePpm((char*)"MatImgCtr","",MatImgCenter,Length,Width);

 //>CenterPrototypeSegmentation
 for(k=0;k<NbImg;k++) VctLabel[k]=0;
 ComputeCtr(VctSeg,NbImg,Length,Width,1,VctLabel,VctCtrSeg,Init);

 CopyMat(MatSegVoIMoy,VctCtrSeg[0],Length,Width);
 SaveImagePgm((char*)"MatSegVoIMoy","",MatSegVoIMoy,Length,Width);
 
 //>Farest_Img_Seg
 MaxVoI=0.0;
 IndxMaxVoI=-1;
 for(k=0;k<NbImg;k++)
   { VoI=VariationInformation(MatSegCenter,VctSeg[k],Length,Width); 
     printf("\n  [%d::%.2f]",k,VoI);
     fflush(stdout);
     VctVoIDist[k]=VoI;
     if (VoI>MaxVoI) { MaxVoI=VoI; IndxMaxVoI=k; printf("*"); } }

 CopyMat(MatSegFarest,VctSeg[IndxMaxVoI],Length,Width);
 SaveImagePgm((char*)"MatSegFar","",MatSegFarest,Length,Width);
 CopyMat(MatImgFarest,VctImgCol[IndxMaxVoI],Length,Width);
 SaveImagePpm((char*)"MatImgFar","",MatImgFarest,Length,Width);

 //>FreeMemory 
 if (VctVoIMoy)  free_fmatrix_1d(VctVoIMoy); 
 if (VctVoIDist) free_fmatrix_1d(VctVoIDist); 
 if (MatSegCenter) free_imatrix_2d(MatSegCenter); 
 if (MatImgCenter) free_imatrix_3d(MatImgCenter,TROIS); 
 if (MatSegFarest) free_imatrix_2d(MatSegFarest); 
 if (MatImgFarest) free_imatrix_3d(MatImgFarest,TROIS); 
 if (MatSegVoIMoy) free_imatrix_2d(MatSegVoIMoy); 
}

//----------------------------------------------------------
// [5]>QueryByDrawing
// -----------------------------------
//
//----------------------------------------------------------
void SCVOIC::QueryByDrawing()
{
  int k,l;
 int IndxMinVoI;;
 float MinVoI;
 
 //>Presentation
 printf("\n  QueryByDrawing [%s][NbNeigbors=%d]",QueryDrawingFileName,NBNEIGHBORS_QBD);   
 printf("\n  ----------------------------------------------");
 
 //>LoadFiles
 LoadFiles();

 //>MemoryAllocation
 float*  VctVoIDist=fmatrix_allocate_1d(NbImg);
 int***  VctMatSegNearest=imatrix_allocate_3d(NBNEIGHBORS_QBD,Length,Width);
 int**** VctMatImgNearest=imatrix_allocate_4d(NBNEIGHBORS_QBD,TROIS,Length,Width);
 int**   QueryDrawing=imatrix_allocate_2d(Length,Width);
 float** QueryDrawingF=fmatrix_allocate_2d(Length,Width);

 //>LoadQueryDrawing
 load_image_pgm(QueryDrawingFileName,QueryDrawing,Length,Width);
 //SaveImagePgm((char*)"ImgDebug","",QueryDrawing,Length,Width);
 //exit(-1);
 CopyMat(QueryDrawingF,QueryDrawing,Length,Width);
 ConvertClassRegion(QueryDrawingF,Length,Width);
 CopyMat(QueryDrawing,QueryDrawingF,Length,Width);

 //>Fill_VctVoIDist
 for(k=0;k<NbImg;k++) 
    { VctVoIDist[k]=VariationInformation(QueryDrawing,VctSeg[k],Length,Width);
      printf("\n  [%d::%.2f]",k,VctVoIDist[k]);
      fflush(stdout); }

 //>Nearest_Img_Seg_
 printf("\n\n  Seeking...");
 for(l=0;l<NBNEIGHBORS_QBD;l++)
    { MinVoI=BIGNUMBER;
      IndxMinVoI=-1;
      for(k=0;k<NbImg;k++) if (VctVoIDist[k]<MinVoI) { MinVoI=VctVoIDist[k]; IndxMinVoI=k; }
     
      printf(" <%d::%.2f>",IndxMinVoI,MinVoI);
      fflush(stdout);
      VctVoIDist[IndxMinVoI]=BIGNUMBER;

      CopyMat(VctMatSegNearest[l],VctSeg[IndxMinVoI],Length,Width);
      CopyMat(VctMatImgNearest[l],VctImgCol[IndxMinVoI],Length,Width); }

 //>Save
 Visu2D_Img(VctMatSegNearest,NBNEIGHBORS_QBD,Length,Width,(char*)"NearestSeg");
 Visu2D_Img(VctMatImgNearest,NBNEIGHBORS_QBD,Length,Width,(char*)"NearestImg");

 //>FreeMemory 
 if (VctVoIDist) free_fmatrix_1d(VctVoIDist); 
 if (VctMatSegNearest) free_imatrix_3d(VctMatSegNearest,NBNEIGHBORS_QBD); 
 if (VctMatImgNearest) free_imatrix_4d(VctMatImgNearest,TROIS,NBNEIGHBORS_QBD); 
 if (QueryDrawing) free_imatrix_2d(QueryDrawing);
 if (QueryDrawingF) free_fmatrix_2d(QueryDrawingF);
}
